import mpi4py
import supernest
import supernest.core
import nestcheck.data_processing
import nestcheck.estimators
import nestcheck.plots

from pypolychord.settings import PolyChordSettings
from pypolychord import run_polychord

import numpy as np
import matplotlib.pyplot as plt

nDims = 2
bounds = np.array([[-30, -10], [30, 10]])
means = np.array([0.2, 0])
covmat = np.array([[0.1, 0.05], [0.05, 0.2]])
proposal_mean = np.array([0.2, 0])
proposal_covmat = np.array([[0.3, 0], [0, 0.4]])


def uniform_with_gaussian_like(nDims, mu, sigma, bounds):
    thetamin = bounds[0]
    thetamax = bounds[1]

    def ll(theta):
        delta = theta - mu
        ll = -np.linalg.slogdet(2 * np.pi * sigma)[1] / 2
        ll -= np.linalg.multi_dot([delta, np.linalg.inv(sigma), delta]) / 2
        return ll, []

    def up(cube):
        return thetamin + cube * (thetamax - thetamin)

    return up, ll


runaway_evidence = False
average_evidence = None


def dumper(live, dead, logweights, logZ, logZerr):
    global average_evidence, runaway_evidence
    if average_evidence is None:
        average_evidence = logZ
    else:
        if logZ - average_evidence > 10:
            print('---------------------------------')


def main():
    hypothesis = uniform_with_gaussian_like(2, means, covmat, bounds)
    proposal = supernest.gaussian_proposal(bounds, proposal_mean,
                                           proposal_covmat, hypothesis[1])
    # superProposal = supernest.superimpose([hypothesis, proposal], nDims)
    settings = PolyChordSettings(nDims, 0)
    settings.read_resume = False
    settings.feedback = 0
    settings.file_root = "super_test"
    # supernest.core.debug = True
    run_polychord(proposal.likelihood,
                  nDims,
                  0,
                  settings,
                  prior=proposal.prior,
                  dumper=dumper)


if __name__ == '__main__':
    main()



"""This is the reference implementation of the superpositional
stochastic mixture, that was used to generate my entire master\'s
thesis. 

Pass any of the mixture Models a list of models, and you\'d get the
proper mixture of the type that you needed. You mostly want to use the
stochastic mixture.

"""

from abc import ABC
from random import random, seed

from numpy import concatenate

from .polychord import Model


def _are_all_elements_identical(lst):
    return not lst or lst.count(lst[0]) == len(lst)


class AbstractMixtureModel(Model, ABC):
    """This is the abstract base class that defines a general mixture. If
    you either have your own implementation of a superpositional
    mixture, or have a different idea entirely, for a mixture of
    several models, you want to subclass this one. 


    Most of the heavy lifiting is figureing out the dimesnionality for
    arbitrary model inputs, and figuring out an efficient log_like and
    prior_quantile.

    """
    default_file_root = 'MixtureModel'

    def __str__(self):
        return f'\nMixture of \n{self.models}'

    def __init__(self, models, file_root=default_file_root, **kwargs):
        self.models = models
        if not _are_all_elements_identical([x.nDims for x in models]):
            raise ValueError(
                "Models in mixture have different dimensionality.")
        self.nDims = max([x.dimensionality for x in models])
        if not _are_all_elements_identical([x.num_derived for x in models]):
            raise ValueError(
                "Models in mixture have different derived parameters.")
        self.nDerived = models[0].num_derived
        super().__init__(self.dimensionality, self.num_derived, file_root, **kwargs)

    def test_quantile(self):
        __doc__ = super().__doc__
        for m in self.models:
            m.test_quantile()
        super().test_quantile()

    def test_log_like(self):
        __doc__ = super().__doc__
        for m in self.models:
            m.test_log_like()
        super().test_log_like()

    @property
    def num_derived(self):
        __doc__ = super().__doc__
        return self.nDerived


class StochasticMixtureModel(AbstractMixtureModel):
    """This is a stochastic mixture model. As described in my Masters'
    thesis also available at https://github.com/appetrosyan/LCDM-NS/

    Initially the `super_nest.superimpose` function was a carbon copy
    of this method. However, I didn't want to impose on the user the
    requirement to keep track of dimensionality, so I decided to cut
    all of the dimensionality management and reserve it for the
    framework proper.

    """
    default_file_root = 'StochasticMixture'

    def __init__(self, models, settings=None, file_root=default_file_root):
        super().__init__(models, file_root=file_root, settings=settings)

    @property
    def dimensionality(self):
        __doc__ = super().dimensionality.__doc__
        return self.nDims + len(self.models)

    def _unpack(self, theta):
        physical_params = theta[:self.nDims]
        choice_probabilities = theta[self.nDims:-1]
        index = theta[-1:].item()
        return physical_params, choice_probabilities, index

    def log_likelihood(self, theta):
        __doc__ = super().__doc__
        t, _, m = self._unpack(theta)
        _current_model = self.models[int(m)]
        _nDims = _current_model.dimensionality
        log_l, phi = _current_model.log_likelihood(t[:_nDims])
        return log_l, phi

    def prior_quantile(self, hypercube):
        __doc__ = super().__doc__ 
        t, b, _ = self._unpack(hypercube)
        norm = b.sum() if b.sum() != 0 else 1
        ps = b / norm
        index = 0
        h = hash(tuple(t))
        seed(h)
        r = random()
        for p in ps:
            if r > p:
                break
            index += 1
        _nDims = self.models[index].dimensionality
        cube, cube_ = t[:_nDims], t[_nDims:]
        theta = self.models[index].prior_quantile(cube)
        return concatenate([theta, cube_, b, [index]])



import supernest
from pypolychord.settings import PolyChordSettings
from pypolychord import run_polychord
import matplotlib.pyplot as plt
import numpy as np
import mpi4py


def uniform_with_gaussian_like(nDims, mu, sigma, bounds):
    thetamin = bounds[0]
    thetamax = bounds[1]
    invCov = np.linalg.inv(sigma)

    def ll(theta):
        return -(theta - mu) @ invCov @ (theta - mu) / 2, []

    def up(cube):
        return thetamin + cube * (thetamax - thetamin)

    return up, ll


def normal_run_settings(nDims,
                        nlive,
                        file_root='control_uniform_gaussian_like'):
    settings = PolyChordSettings(nDims, 0)
    settings.read_resume = False
    settings.feedback = 0
    settings.nlive = nlive
    settings.file_root = file_root

    return settings


def test_uniform(nDims, mu, sigma, bounds, nlive):
    prior, like = uniform_with_gaussian_like(nDims, mu, sigma, bounds)

    settings = normal_run_settings(nDims, nlive)
    output = run_polychord(like, nDims, 0, settings, prior)
    return output, prior, like


def test_proposal(nDims, mu, sigma, bounds, nlive):
    prior, like = uniform_with_gaussian_like(nDims, mu, sigma, bounds)

    pp, ll = supernest.gaussian_proposal(bounds, mu, sigma, loglike=like)
    settings = normal_run_settings(nDims,
                                   nlive,
                                   file_root="supernest_proposal")
    output = run_polychord(ll, nDims, 0, settings, pp)
    return output, pp, ll


def test_supernest(nDims, mu, sigma, bounds, nlive):
    prior, like = uniform_with_gaussian_like(nDims, mu, sigma, bounds)

    pp, ll = supernest.truncated_gaussian_proposal(bounds,
                                                   mu,
                                                   sigma,
                                                   loglike=like)
    dims, ppp, lll = supernest.superimpose([(prior, like), (pp, ll)], nDims)

    settings = normal_run_settings(dims,
                                   nlive,
                                   file_root="supernest_superimposed")

    output = run_polychord(lll, dims, 0, settings, ppp)
    return output, ppp, lll


def deltas(ends, mu, sigma, nlive):
    retZ = []
    retZerr = []
    for end in ends:
        args = truncate(mu, sigma, bounds, end)
        args.append(nlive)
        uni = test_uniform(*args)[0]
        pro = test_proposal(*args)[0]
        retZ.append(uni.logZ - pro.logZ)
        retZerr.append(max(uni.logZerr, pro.logZerr))
    return retZ, retZerr


def truncate(mu, sigma, bounds, end):
    _mu = mu[:end]
    _sigma = sigma[:end, :end]
    _bounds = bounds[:, :end]
    _bounds = _bounds
    return [len(_mu), _mu, _sigma, _bounds]


def main(nlive=30, end=27):
    global uniform, proposal, x, y, yerr, mu, sigma, bounds, args
    mu = np.loadtxt(
        '/home/app/Git/sspr/supernest/tests/cosmology_data/means.npy')
    sigma = np.loadtxt(
        '/home/app/Git/sspr/supernest/tests/cosmology_data/covmat.npy')
    bounds = np.loadtxt(
        '/home/app/Git/sspr/supernest/tests/cosmology_data/bounds.npy')
    bounds = bounds.T

    x = range(3, 27)
    y, yerr = deltas(x, mu, sigma, nlive)
    np.savetxt('deltas.npy', y)


if __name__ == '__main__':
    main()



import supernest
from pypolychord.settings import PolyChordSettings
from pypolychord import run_polychord
import scipy.stats as st
import matplotlib.pyplot as plt
import numpy as np
import mpi4py


def uniform_with_gaussian_like(nDims, mu, sigma, bounds):
    thetamin = bounds[0]
    thetamax = bounds[1]
    invCov = np.linalg.inv(sigma)

    def ll(theta):
        return -(theta - mu) @ invCov @ (theta - mu) / 2, []

    def up(cube):
        return thetamin + cube * (thetamax - thetamin)

    return up, ll


def normal_run_settings(nDims,
                        nlive,
                        file_root='control_uniform_gaussian_like'):
    settings = PolyChordSettings(nDims, 0)
    settings.read_resume = False
    settings.feedback = 0
    settings.nlive = nlive
    settings.file_root = file_root

    return settings


def test_uniform(nDims, mu, sigma, bounds, nlive):
    prior, like = uniform_with_gaussian_like(nDims, mu, sigma, bounds)

    settings = normal_run_settings(nDims, nlive)
    output = run_polychord(like, nDims, 0, settings, prior)
    return output, prior, like


def test_proposal(nDims, mu, sigma, bounds, nlive):
    prior, like = uniform_with_gaussian_like(nDims, mu, sigma, bounds)

    pp, ll = supernest.truncated_gaussian_proposal(bounds,
                                                   mu,
                                                   sigma,
                                                   loglike=like)
    settings = normal_run_settings(nDims,
                                   nlive,
                                   file_root="supernest_proposal")
    output = run_polychord(ll, nDims, 0, settings, pp)
    return output, pp, ll


def test_supernest(nDims, mu, sigma, bounds, nlive):
    prior, like = uniform_with_gaussian_like(nDims, mu, sigma, bounds)
    pp, ll = supernest.truncated_gaussian_proposal(bounds,
                                                   mu,
                                                   sigma,
                                                   loglike=like)
    dims, ppp, lll = supernest.superimpose([(prior, like), (pp, ll)], nDims)

    settings = normal_run_settings(dims,
                                   nlive,
                                   file_root="supernest_superimposed")

    output = run_polychord(lll, dims, 0, settings, ppp)
    return output, ppp, lll


def deltas(boundaries, nDims, mu, sigma, nlive):
    retZ = []
    retZerr = []
    for b in boundaries:
        bounds = (-b, b)
        uni = test_uniform(nDims, mu, sigma, bounds, nlive)[0]
        # pro = test_supernest(nDims, mu, sigma, bounds, nlive)[0]
        pro = test_proposal(nDims, mu, sigma, bounds, nlive)[0]
        retZ.append(uni.logZ - pro.logZ)
        retZerr.append(max(uni.logZerr, pro.logZerr))
    return retZ, retZerr


def test_truncated_gaussian(boundaries, nDims, mu, sigma):
    pp, _ = supernest.truncated_gaussian_proposal(boundaries,
                                                  mu,
                                                  sigma,
                                                  bounded=False)
    xs = np.linspace(0, 1, 100000)
    ys = np.array([pp(x) for x in xs])
    print(ys)
    yys = np.array([
        st.truncnorm.ppf(x, boundaries[0], boundaries[1], scale=sigma)
        for x in xs
    ])
    plt.figure()
    plt.hist(yys, alpha=0.8, label='scipy', bins=100)
    plt.hist(ys, alpha=0.8, label='supernest', bins=100)
    plt.legend()
    plt.show()


def main(a=1, it=5):
    global uniform, proposal, delta, x, deltaerr
    nlive = 120
    nDims = 2
    mu = np.array([0, 1])
    sigma = np.array([[1, 0], [0, 1]])
    bounds = (np.array([-a]), np.array([a]))
    x = np.linspace(0.001 * a, a, it)
    delta, deltaerr = deltas(x, nDims, mu, sigma, nlive)
    # test_truncated_gaussian(bounds, nDims, mu, sigma)


if __name__ == "__main__":
    main()



r"""Superpositional model repartitiong for accelerated nested sampling.

model repartitioning also known as posterior or more accurrately
conclusion re-partitioning is a technique that allows reshaping the
input functions: prior and likelihood and achieving faster Bayesian
inference by way of Nested Sampling.

IMPORTANT: This is an early version of the software, which does not
yet have the links to the proper publication, nor indeed is guaranteed
to function. If you use this for your research, please consider
waiting until the full release.

"""
from .core import superimpose 
from .core import gaussian_proposal 
from .core import truncated_gaussian_proposal
from .core import Proposal, NDProposal



import unittest
import numpy as np
import supernest as sn


class TestSupernestExceptions(unittest.TestCase):
    def test_bounds_transposed_is_caught(self):
        mu = np.array([1, 2, 3])
        sigma = np.array([[1, 2, 3], [3, 6, 8], [4, 5, 3]])
        bounds = np.array([[1, 2, 3], [4, 5, 6]]).T
        self.assertTrue(len(mu == len(sigma)))
        self.assertTrue(len(mu == len(sigma[0])))
        self.assertWarnsRegex(UserWarning, 'Bounds should be transposed.',
                              sn.gaussian_proposal, *[bounds, mu, sigma])

    def test_lopsided_bounds(self):
        mu = np.array([1, 2, 3])
        sigma = np.array([[1, 2, 3], [3, 6, 8], [4, 5, 3]])
        bounds = np.array([[1, 2, 3], [4, 5, 6, 7]], dtype=object)
        self.assertTrue(len(bounds[0]) != len(bounds[1]))
        self.assertRaisesRegex(ValueError, 'Lopsided bounds',
                               sn.gaussian_proposal, *[bounds, mu, sigma])

    def test_incompatible_bounds(self):
        mu = np.array([1, 2, 3])
        sigma = np.array([[1, 2, 3], [3, 6, 8], [4, 5, 3]])
        bounds = np.array([[1, 2, 3, 4], [4, 5, 6, 7]], dtype=object)
        self.assertTrue(len(bounds[0]) == len(bounds[1]))
        self.assertTrue(len(bounds[0]) != len(mu))
        self.assertRaisesRegex(
            ValueError,
            'Proposal mean and boundaries are of imcompatible length',
            sn.gaussian_proposal, *[bounds, mu, sigma])

    def test_mu_doesnt_match_cov(self):
        mu = np.array([1, 2, 3, 4])
        sigma = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        bounds = (np.array([0, 1, 2, 3]), np.array([-0, -1, -2, -3]))
        self.assertTrue(len(mu) != len(sigma))
        self.assertRaises(ValueError, sn.gaussian_proposal,
                          *[bounds, mu, sigma])
        sigma2 = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        self.assertTrue(len(mu) == len(sigma2))
        self.assertTrue(len(mu) != len(sigma2[0]))
        self.assertRaisesRegex(ValueError, 'Dimensions of covariance and mean',
                               sn.gaussian_proposal, *[bounds, mu, sigma2])


if __name__ == "__main__":
    unittest.main()



"""This modules is for internal use, mostly. THis file describes the
base API that the entire framework follows. Changes here will be felt
in every single subclass of the framework, so changing things here,
should almost never alter the expected behaviour, but rather only the
internal behaviour, and handle the bugs.


If you are writing your own code, and want to make use of the sanity
checks, the somewhat less-than-elegant structure, use this file as a
last resort. For subclassing, it's much easier to look at the
subclasses in ParameterCovarianceModel. Specifically, there you shall
find the abstract base class that tells you what to do with the
likelihood, and a few examples of how to encode a prior other than
uniform.

The file itself is structured as a tutorial (more-less).
"""
from copy import deepcopy

from anesthetic import NestedSamples
from numpy import zeros

# As of now PolyChord is not `pip install pypolychord` -able
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from pypolychord import run_polychord
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from pypolychord.settings import PolyChordSettings


class Model:
    """A Base class for the models in the `super_nest` framework.

    This is what you need to sbuclass if you want to do custom stuff
    using PolyChord.

    When Sub-classing you should do three things. You need to define a
    constructor to keep the local data that you may need for
    loglikelihood and prior_quantile.

    Define the prior and the likelihood using the two unimplemented
    methods.

    """
    default_file_root = 'blankModel'

    def __init__(self, dimensionality, number_derived, file_root='', **kwargs):
        self.settings = PolyChordSettings(dimensionality, number_derived)
        self.settings.file_root = file_root

    def log_likelihood(self, theta):
        """A Ln(likelihood) of the theta given the model. With a uniform
        prior, this defines the posterior distribution up to a
        multiplicative factor.

        Parameters
        ----------
        theta : array(self.dimensionality, dtype=numpy.float64)
            Physical parameters' values.

        Returns
        -------
        logL: numpy.float64, [derived..] : list(num_derived, np.float64))

            The first element is the log-likelihood, the second is a list
            of the derived parameters.

        """
        raise NotImplementedError()

    def prior_quantile(self, hypercube):
        """Inverse Cumulative distribution function of the prior. Aka the
        quantile. If the prior has PDF \\pi, then this is (CDF (\\pi))^-1.

        Parameters
        ----------

        hypercube: array(self.dimensionality, dtype=numpy.float64)
            Physical parameters' images in a unit hypercube, where their
            distribution is uniform.

        Returns
        -------
        theta: array(self.dimensionality, dtype=numpy.float64)
            Physical parameters.
        """
        raise NotImplementedError()

    @property
    def dimensionality(self):
        """This is the length of the physical parameter vector to be used. You
        don't need to override this, unless you know what you're doing, and
        what you're doing involves passing extra dummy parameters to the
        sampler. This value is used internally for consistency checking, so
        make sure that it's correct.

        """
        return int(self.nDims)

    @property
    def num_derived(self):
        """The number of derived parameters. Override if you have derived
        parameters in self.log_likelihood.

        """
        return 0

    def test_log_like(self):
        """Not a user facing function. This is run before nested sampling is
        executed, so you should put all the sanity checking code that requires
        the self.log_likelihood here.

        """
        p = self.log_likelihood(zeros(self.dimensionality))
        try:
            _, _ = p
        except ValueError as e:
            raise ValueError(
                e.msg + "Did you forget to return the derived parameters?")

    def test_quantile(self):
        """Not a user facing function. This is run before nested sampling is
        executed, so you should put all the sanity checking code that requires
        the self.prior_quantile here.

        """
        _nDims = len(self.prior_quantile(zeros(self.dimensionality)))
        if _nDims != self.dimensionality:
            raise ValueError(
                f'Prior has the wrong dimensions: expect {_nDims}'
                f'vs actual {self.dimensionality}')

    def nested_sample(self, **kwargs):
        """A safer and more configurable way of running the `PyPolyChord`
        nested sampler.

        This will raise errors if you passed in inconsistent
        dimesnions.

        Parameters
        ----------

        **kwargs: dict
        Options that pypolychord.settings.PolyChordSettings object would accept.

        """
        self.test_log_like()
        self.test_quantile()
        _settings = self.setup_settings(**kwargs)
        output = run_polychord(self.log_likelihood, self.dimensionality,
                               self.num_derived, _settings, self.prior_quantile)
        try:
            samples = NestedSamples(
                root=f'./chains/{_settings.file_root}')
        except ValueError as e:
            print(e)
            samples = None
        return output, samples

    # noinspection SpellCheckingInspection
    def setup_settings(self, file_root=None,
                       live_points=175, resume=True, verbosity=0):
        """This is a helper function that sets PolyChord up with sane defaults.
        """
        _settings = deepcopy(self.settings)
        _settings.feedback = verbosity
        if file_root is not None:
            _settings.file_root = file_root
        _settings.read_resume = resume
        _settings.nlive = live_points
        return _settings



"""This classs provides a convenient wrapper around the standard
models, and allows introducing an offset between the likelihood and
prior coordinate systems. This is useful for testing and simulating
priors that have are at variance with the likelihood. You only need
this, if you want to have a prior that\'s **deliberately** offset from
where you expect the likelihood peak to be located at.

"""
from numpy import pad

from .polychord import Model


class OffsetModel(Model):
    """An abstract offset model decorator. Pass it any model and it will
    return a model that's offset by a vector.

    """
    default_file_root = 'OffsetModel'

    def __str__(self):
        return f'{self.model.__str__()}\nOffset by {self.offset}'

    def __repr__(self):
        return f'{self.model.__repr__()} Offset by {self.offset}'

    def __init__(self, base_model, offset, file_root=default_file_root, **kwargs):
        self.model = base_model
        self.offset = offset
        if not self.offset.size == self.model.dimensionality:
            self.offset = pad(self.offset, (0, abs(
                self.model.dimensionality - self.offset.size)))
        super().__init__(base_model.dimensionality,
                         base_model.num_derived, file_root=file_root, **kwargs)

    def log_likelihood(self, theta):
        return self.model.log_likelihood(theta - self.offset)

    def prior_quantile(self, *args):
        return self.model.prior_quantile(*args)

    @property
    def dimensionality(self):
        return self.model.dimensionality



"""This is an object-oriented interface to proposals and PolyChord.

The intent is to ease exploration of models with plain PolyChord (and
anesthetic). This is here to help propototype the types of proposals,
likelihood functions and calibrate the Bayesian inference.

This is less helpful in cases where you want to adjust a project
that's already using Bayesian inference. For the purposes of **just**
using superpositional mixtures, refer to `super_nest.superimpose()`.

Most of the prototyping is usually done with a lightweight likelihood
function that's ordinarily well-approximated by a Gaussian (or a sum
of gaussians). For this purpose see the submodule

`gaussian_models`

If you need to write custom code, the functions in the same module
provide a good example of how to subclass the Model, and therefore
adapt the interface to your workflow.
"""
from .polychord import Model



r"""Module containing superimpose and a gaussian proposal.

The usage is as follows. Normally the proposal is approximated by a
correlated Gaussian distribution. We (for now) approximate that
further to a spherically symmetric gaussian and use that as the guide
for nested sampling.

Afterwards, the important step is to put the Gaussian proposal into a
superpositional mixture.  This is done via a functional interface for
ease and portability.

"""
import random
import numpy as np
import numpy.linalg
from scipy.special import erf, erfinv
from collections import namedtuple
import warnings

Proposal = namedtuple("Proposal", ['prior', 'likelihood'])
NDProposal = namedtuple("NDProposal", ['nDims', 'prior', 'likelihood'])

debug = False


def superimpose(models: list, nDims: int = None):
    r"""Superimpose functions for use in nested sampling packages.

    Parameters
    ----------
    models: list(tuples(callable, callable))

    This is a list of pairs of functions. The first functions
    (quantile-like) will be interpreted as prior quantiles. They will
    be made to accept extra arguments from the hypercube, and produce
    extra parameters as output.

    The secondary function will be made to accept extra parameters,
    and ignore all but the last parameter. The functions need to be a
    consistent partitioning of the model, as described in the
    stochastic superpositional mixing paper in mnras.

    In short, if the prior spaces to which the two functions coorepond
    is the same, for all functions, you only need to make sure that
    the product of the prior pdf and the likelihood pdf is the same
    acroos different elemtns of the tuple. If they are not the same,
    you must make sure that the integral of their product over each
    prior space is the same, and that the points which correspond to
    the same locations in the hypercube align.

    nDims=None: int
    Optionally, if you want to have `superimpose`
    produce a number of dimensions for use with e.g. PolyChord, and to
    guard againt changes in the calling conventions and API, just pass
    the nDims that you would pass to PolyChord.Settings, and the
    run_polychord function.


    Returns
    -------
    (prior_quantile: callable, likelihood: callable) : tuple
    if nDims is None,
    returns a tuple of functions: the superposition of the prior
    quantiles and the likelihoods (in that order).

    (nDims: int, prior_quantile: callable, likelihood: callable): tuple
    if the optional argument nDims is not None, the output also
    contains an nDims: the number of dimensions that you should ask
    your dimesnional sampler to pass.

    """
    priors, likes = [p for p, _ in models], [l for _, l in models]

    def prior_quantile(cube):
        physical_params = cube[:-len(models)]
        choice_params = cube[-len(models):-1]
        index = 0
        norm = choice_params.sum()
        norm = 1 if norm == 0 or len(choice_params) == 1 else norm
        probs = choice_params / norm
        h = hash(tuple(physical_params))
        random.seed(h)
        rand = random.random()
        for p in probs:
            if rand > p:
                break
            index += 1
        theta = priors[index](physical_params)
        ret = np.array(np.concatenate([theta, probs, [index]]))
        return ret

    def likelihood(theta):
        try:
            physical_params = theta[:-len(models)]
        except SystemError:
            warnings.warn(f'theta = {theta} {theta[:-len(models)]}')
            physical_params = theta[:-len(models)]
        index = int(theta[-1:].item())
        ret = likes[index](physical_params)
        return ret

    if nDims is not None:
        return NDProposal(nDims + len(models), prior_quantile, likelihood)
    else:
        return Proposal(prior_quantile, likelihood)


def _eitheriter(ab):
    a, b = ab
    return hasattr(a, '__iter__') or hasattr(b, '__iterb__')


def __snap_to_edges(cube, theta, a, b):
    ret = theta
    if np.any(np.isclose(cube, 0)) or np.any(np.isclose(cube, 1)):
        for i in range(len(cube)):
            if np.isclose(cube[i], 0):
                ret[i] = a[i] if hasattr(a, '__iter__') else a
            elif np.isclose(cube[i], 1):
                ret[i] = b[i] if hasattr(b, '__iter__') else b
            else:
                pass
    return ret


def __guard_against_inf_nan(cube, theta, logzero, loginf):
    ret = theta
    if not np.all(np.isfinite(ret)):
        for i in range(len(cube)):
            if np.isclose(cube[i], 0):
                ret[i] = logzero
            elif np.isclose(cube[i], 1):
                ret[i] = loginf
            elif not np.isfinite(ret[i]):
                ret[i] = logzero
            else:
                pass
    else:
        return ret


def gaussian_proposal(bounds: np.ndarray,
                      mean: np.ndarray,
                      covmat: np.ndarray,
                      loglike: callable = None):
    r"""Produce a Gaussian proposal.

    Given a uniform prior defined by bounds, produces the corrected
    loglikelihood and prior.

    Parameters
    ----------
    bounds: array-like
        A tuple-like or array-like that contains the (min, max) of the
        original uniform prior.

    mean: array-like
        A vector of the means of the gaussian approximation of the proposal

    covmat: array-like
        A matrix containing the covariance of the gaussian proposal.

    loglike: callable (optional)
        The loglikelihood function of the original model to be corrected.

    Returns
    -------
    proposal: Proposal (tuple(prior, loglike))
    """
    covmat, a, b = _process_stdev(covmat, mean, bounds)
    log_box = np.log(b - a).sum() if _eitheriter(
        (a, b)) else len(mean) * np.log(b - a)
    log_box = -log_box
    try:
        invCov = np.linalg.inv(covmat)
    except numpy.linalg.LinAlgError:
        invCov = 1 / covmat

    def __quantile(cube):
        theta = np.sqrt(2) * erfinv(2 * cube - 1)
        theta = mean + covmat @ theta
        return theta

    def __correction(theta):
        if loglike is None:
            ll, phi = 0, []
        else:
            ll, phi = loglike(theta)
        corr = -numpy.linalg.multi_dot([(theta - mean), invCov,
                                        (theta - mean)]) / 2
        corr -= np.log(
            2 * np.pi) * len(mean) / 2 + np.linalg.slogdet(covmat)[1] / 2
        if debug:
            print(f'll={ll}\t corr={corr}')
        return (ll - corr + log_box), phi

    return Proposal(__quantile, __correction)


def truncated_gaussian_proposal(bounds: np.ndarray,
                                mean: np.ndarray,
                                stdev: np.ndarray,
                                loglike: callable = None):
    r"""Produce a truncated Gaussian proposal.

    Given a uniform prior defined by bounds, it produces a gaussian
    prior quantile and a correction to the log-likelihood.

    Parameters
    ----------
    bounds : array-like
        A tuple with bounds of the original uniform prior.

    mean : array-like
        The vector \mu at which the proposal is to be centered.

    stdev : array-like
        The vector of standard deviations. Currently only
        uncorrelated Gaussians are supported.

    loglike: callable: (array-like) -> (real, array-like), optional
        The callable that constitutes the model likelihood.  If provided
        will be included in the output. Otherwise assumed to be
        lambda () -> 0


    Returns
    -------
    (prior_quantile, loglike_corrected): tuple(callable, callable)
    This is the output to be used in the stochastic mixing. You can
    use it directly, if you\'re certain that this is the exact shape of
    the posterior. Any deviation, however, will be strongly imprinted
    in the posterior, so you should think carefully before doing this.

    """
    stdev, a, b = _process_stdev(stdev, mean, bounds)
    # truncation requires the covmat to be diagonal
    try:
        stdev = stdev.diagonal()
    except ValueError:
        warnings.warn(f'stdev={stdev} couldn\'t be diagonalised')
    log_box = np.log(b - a).sum() if _eitheriter(
        (a, b)) else len(mean) * np.log(b - a)
    log_box = -log_box

    # Convenice variable to avoid duplicating code
    RT2, RTG = np.sqrt(2), np.sqrt(1 / 2) / stdev
    da = erf((a - mean) * RTG)
    db = erf((b - mean) * RTG)

    def __quantile(cube):
        theta = RT2 * erfinv((1 - cube) * da + cube * db)
        theta = mean + stdev * theta
        theta = __snap_to_edges(cube, theta, a, b)
        return theta

    def __correction(theta):
        if loglike is None:
            ll, phi = 0, []
        else:
            ll, phi = loglike(theta)
        corr = -((theta - mean)**2) / (2 * stdev)
        corr -= np.log(2 * np.pi * stdev**2) / 2
        corr -= np.log((db - da) / 2)
        corr = corr.sum()
        if debug:
            print(f'll={ll}\tcorr={corr}\tlog_box={log_box}')
        return (ll - corr + log_box), phi

    return Proposal(__quantile, __correction)


def _process_stdev(stdev, mean, bounds):
    if isinstance(stdev, float):
        stdev = np.zeros(len(mean)) + stdev
    elif len(mean) != len(stdev):
        raise ValueError(
            'Proposal Mean and covariance are of incompatible lengths: ' +
            f'len(mean)={len(mean)} vs. len(stdev)={len(stdev)}')
    else:
        if len(stdev[0]) != len(mean):
            raise ValueError(
                'Dimensions of covariance and mean don\'t match' +
                f'len(stdev)={len(stdev)} vs len(mean)={len(mean)}')

    try:
        a, b = bounds
    except ValueError:
        a, b = bounds.T
        warnings.warn('Bounds should be transposed.')

    if isinstance(a, (int, float)) and isinstance(b, (int, float)):
        pass
    else:
        if len(a) != len(b):
            raise ValueError('Lopsided bounds: ' +
                             f'len(a)={len(a)} vs. len(b)={len(b)}')

        if 0 != len(a) != len(mean):
            raise ValueError(
                'Proposal mean and boundaries are of imcompatible lengths: ' +
                f'len(a)={len(a)} vs len(mean)={len(mean)}')

    return stdev, a, b



"""Setup code for the SSPR Package."""
import setuptools
import os

SHORT_DESCRIPTION = 'A wrapper for use of SSPR in \
nested sampling packages such as PolyChord and Multinest'
LONG_DESCRIPTION = SHORT_DESCRIPTION

with open("./README.md") as readme:
    LONG_DESCRIPTION = readme.read()

version = "3.1.2"

if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG']

setuptools.setup(
    name='supernest',
    version=version,
    author='Aleksandr Petrosyan',
    author_email='a-p-petrosyan@yandex.ru',
    description=SHORT_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/a-p-petrosyan/sspr',
    install_requires=['anesthetic', 'numpy', 'matplotlib'],
    packages=setuptools.find_packages(),
    license='LGPLv3',
    python_requires='>=3.6',
    test_suite="tests",
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)'
    ]
)



import unittest
import supernest as sn
import numpy as np


class TestTyping(unittest.TestCase):
    def test_proposal_is_proper_return_type(self):
        bounds = (0, 1)
        mean = np.array([0])
        stdev = np.array([[1]])
        proposal = sn.gaussian_proposal(bounds, mean, stdev)
        self.assertIsInstance(proposal, sn.Proposal)
        self.assertIsInstance(sn.superimpose([proposal, proposal]),
                              sn.Proposal)
        self.assertIsInstance(sn.superimpose([proposal, proposal], nDims=1),
                              sn.NDProposal)


if __name__ == '__main__':
    unittest.main()



"""This module provides convenience classes that should make writing
your own proposals and using stochastic mixtures much easier. You
don't need to worry about the logistics of truncation, of the problems
that it might cause or indeed of any of the issues caused by passing
in arrays of a different size than nDims.
"""
from abc import ABC

from numpy import pi, array, log, concatenate, diag, sqrt, nextafter
from numpy.linalg import slogdet, multi_dot, inv, pinv
from scipy.special import erf, erfinv
from functools import lru_cache

from .polychord import Model
from pypolychord.priors import UniformPrior


class ParameterCovarianceModel(Model, ABC):
    """
    This is the abstract base class that you would need to sub-class in
    order to produce a ParameterCovarianceModel. This means that you
    provide it with a mean and a covariance and it then produces the
    standard loglikelihood that corresponds to a gaussian. When
    sublassing, you should worry about the correct dimensionality, and
    produce a prior quantile that correponds to the model you want.
    """

    def __str__(self):
        return f'Gaussian posterior with \nmu = {self.mu}\ncov = {self.cov}'

    def __repr__(self):
        return self.__str__().replace('\n', ' ')

    def __init__(self, bounds, mu, cov, file_root='paramCovModel', **kwargs):
        self.a, self.b = bounds
        self.mu = mu
        self.cov = cov
        try:
            self.nDims = mu.size
        except AttributeError:
            self.mu = array(mu)
            self.nDims = mu.size

        try:
            rows, cols = self.cov.shape
        except AttributeError:
            self.cov = array(self.cov)
            rows, cols = self.cov.shape  # raise VE if too few to unpack
        if rows != cols or rows != self.nDims:
            raise ValueError('Dimensions of cov and mean are incompatible: mean – {}, cov ({}, {}) '.format(
                self.nDims, rows, cols))
        try:
            self._invCov = inv(self.cov)
        except:
            print(
                "Singular matrix, reverting to Penrose-Moore inverse, Singular Value Decomposition. ")
            self._invCov = pinv(self.cov)
        super().__init__(self.dimensionality, self.num_derived, file_root, **kwargs)

    def log_likelihood(self, theta):
        """A Ln(likelihood) of the theta given the model. With a uniform
        prior, this defines the posterior distribution up to a
        multiplicative factor. This is a Gaussian Log-likelihood. 

        Parameters
        ----------
        theta : array(self.dimensionality, dtype=numpy.float64)
            Physical parameters' values.

        Returns
        -------
        logL, [derived..] : (log(likelihood): numpy.float64, list(self.num_derived, dtype=numpy.float64))

            The first element is the log-likelihood, the second is a list
            of the derived parameters. 

        """
        delta = theta - self.mu
        ll = - slogdet(2 * pi * self.cov)[1] / 2
        ll -= multi_dot([delta, self._invCov, delta]) / 2
        return ll, []


class PowerPosteriorPrior(ParameterCovarianceModel):
    default_file_root = 'PowerPosteriorModel'
    # Smallest representable +ve float64
    beta_min, beta_max = (nextafter(0, 1), 1)

    def log_likelihood(self, theta):
        """A Ln(likelihood) of the theta given the model. With a uniform
        prior, this defines the posterior distribution up to a
        multiplicative factor. This is a Gaussian-log-likelihood.

        Parameters
        ----------
        theta : array(self.dimensionality, dtype=numpy.float64)
            Physical parameters' values.

        Returns
        -------
        logL, [derived..] : (log(likelihood): numpy.float64, list(self.num_derived, dtype=numpy.float64))

            The first element is the log-likelihood, the second is a list
            of the derived parameters. 

        """
        t = theta[:self.nDims]
        beta = theta[-1]
        log_l, phi = super().log_likelihood(t)
        log_l += log_likelihood_correction(self, beta, t)
        return log_l, phi

    def prior_quantile(self, cube):
        beta = self.beta_min + (self.beta_max - self.beta_min) * cube[-1]
        theta = power_gaussian_quantile(self, cube[:self.nDims], beta)
        return concatenate([theta, [beta]])

    @property
    def dimensionality(self):
        """Dimesnionality of the power posterior model is nDims + 1. 
        """
        return self.nDims + 1


def _erf_term(d, b, g):
    @lru_cache(maxsize=2)
    def helper(t_delta, t_beta, t_sigma):
        hd, hg = array(t_delta), array(t_sigma)
        return erf(hd * sqrt(t_beta / 2) / hg)

    return helper(tuple(d), b, tuple(g))


def power_gaussian_quantile(m, cube, beta=1):
    sigma = diag(m.cov)
    da = _erf_term(m.a - m.mu, beta, sigma)
    db = _erf_term(m.b - m.mu, beta, sigma)
    ret = erfinv((1 - cube) * da + cube * db)
    return m.mu + sqrt(2 / beta) * sigma * ret


def log_box(m):
    if hasattr(m.b, '__iter__') or hasattr(m.a, '__iter__'):
        return log(m.b - m.a).sum()
    else:
        return m.nDims * log(m.b - m.a)


def log_likelihood_correction(model, beta, theta):
    ll = 0

    def ln_z(m, t, b):
        sigma = diag(m.cov)
        ret = - b * (t - m.mu) ** 2 / 2 / sigma ** 2
        ret -= log(pi * sigma ** 2 / 2 / b) / 2
        db = _erf_term(m.b - m.mu, b, sigma)
        da = _erf_term(m.a - m.mu, b, sigma)
        ret -= log(db - da)
        return ret

    ll -= log_box(model)
    ll -= ln_z(model, theta, beta).sum()

    return ll


class GaussianPeakedPrior(ParameterCovarianceModel):
    """This is implemented as a special case of the power posterior
    repartitioning.  One does need to care if it's the right
    function.

    """
    default_file_root = 'GaussianPosteriorModel'

    def log_likelihood(self, theta):
        """A Ln(likelihood) of the theta given the model. With a uniform
        prior, this defines the posterior distribution up to a
        multiplicative factor.

        Parameters
        ----------
        theta : array(self.dimensionality, dtype=numpy.float64)
            Physical parameters' values.

        Returns
        -------
        logL, [derived..] : (log(likelihood): numpy.float64, list(self.num_derived, dtype=numpy.float64))

            The first element is the log-likelihood, the second is a list
            of the derived parameters. 

        """
        log_l, phi = super().log_likelihood(theta)
        log_l += log_likelihood_correction(self, beta=1, theta=theta)
        return log_l, phi

    def prior_quantile(self, cube):
        """Inverse Cumulative distribution function of the prior. Aka the
        quantile. If the prior has PDF \\pi, then this is (CDF (\\pi))^-1.

        Parameters
        ----------

        hypercube: array(self.dimensionality, dtype=numpy.float64) 
            Physical parameters' images in a unit hypercube, where their
            distribution is uniform.

        Returns
        -------
        theta: array(self.dimensionality, dtype=numpy.float64)
            Physical parameters. 
        """
        return power_gaussian_quantile(self, cube)

    @property
    def dimensionality(self):
        """The dimensionality of the truncated gaussians is nDims. 
        """
        return self.nDims


class BoxUniformPrior(ParameterCovarianceModel):
    default_file_root = 'boxUniform'

    def prior_quantile(self, hypercube):
        """Inverse Cumulative distribution function of the prior. Aka the
        quantile. If the prior has PDF \\pi, then this is (CDF (\\pi))^-1.

        Parameters
        ----------

        hypercube: array(self.dimensionality, dtype=numpy.float64) 
            Physical parameters' images in a unit hypercube, where their
            distribution is uniform.

        Returns
        -------
        theta: array(self.dimensionality, dtype=numpy.float64)
            Physical parameters. 
        """
        return UniformPrior(self.a, self.b)(hypercube)


class ResizeablePrior(ParameterCovarianceModel):
    default_file_root = 'ResizeableBoxUniform'
    beta_min = nextafter(0, 1)  # Smallest representable +ve float
    beta_max = 1

    def log_likelihood(self, theta):
        """A Ln(likelihood) of the theta given the model. With a uniform
        prior, this defines the posterior distribution up to a
        multiplicative factor.

        Parameters
        ----------
        theta : array(self.dimensionality, dtype=numpy.float64)
            Physical parameters' values.

        Returns
        -------
        logL, [derived..] : (log(likelihood): numpy.float64, list(self.num_derived, dtype=numpy.float64))

            The first element is the log-likelihood, the second is a list
            of the derived parameters. 

        """
        t = theta[:self.nDims]
        beta = theta[-1]
        if beta <= self.beta_min:
            beta = self.beta_min
        log_l, phi = super().log_likelihood(t / beta)
        log_l += 2 * self.nDims * (log(beta))
        log_l -= log_box(self)
        return log_l, phi

    def prior_quantile(self, hypercube):
        """Inverse Cumulative distribution function of the prior. Aka the
        quantile. If the prior has PDF \\pi, then this is (CDF (\\pi))^-1.

        Parameters
        ----------

        hypercube: array(self.dimensionality, dtype=numpy.float64)
            Physical parameters' images in a unit hypercube, where their
            distribution is uniform.

        Returns
        -------
        theta: array(self.dimensionality, dtype=numpy.float64)
            Physical parameters.
        """
        beta = hypercube[-1:].item()
        # PolyChord refers to Quantile functions as priors.
        # This is not incorrect, but can be confusing.
        uniform = UniformPrior(self.a * beta, self.b * beta)(hypercube[:-1])
        return concatenate([uniform, [beta]])

    @property
    def dimensionality(self):
        """The dimensionality of the resizeable bounds uniform prior is nDims
        + 1.

        """
        return self.nDims + 1
